@@ -10,6 +10,7 @@ import (
 
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/tsdb"
+	"github.com/influxdata/influxql"
 	"github.com/uber-go/zap"
 )
 
@@ -118,7 +119,7 @@ func (e *entry) deduplicate() {
 	e.mu.Lock()
 	defer e.mu.Unlock()
 
-	if len(e.values) == 0 {
+	if len(e.values) <= 1 {
 		return
 	}
 	e.values = e.values.Deduplicate()
@@ -135,6 +136,9 @@ func (e *entry) count() int {
 // filter removes all values with timestamps between min and max inclusive.
 func (e *entry) filter(min, max int64) {
 	e.mu.Lock()
+	if len(e.values) > 1 {
+		e.values = e.values.Deduplicate()
+	}
 	e.values = e.values.Exclude(min, max)
 	e.mu.Unlock()
 }
@@ -147,6 +151,13 @@ func (e *entry) size() int {
 	return sz
 }
 
+// InfluxQLType returns for the entry the data type of its values.
+func (e *entry) InfluxQLType() (influxql.DataType, error) {
+	e.mu.RLock()
+	defer e.mu.RUnlock()
+	return e.values.InfluxQLType()
+}
+
 // Statistics gathered by the Cache.
 const (
 	// levels - point in time measures
@@ -168,14 +179,16 @@ const (
 
 // storer is the interface that descibes a cache's store.
 type storer interface {
-	entry(key string) (*entry, bool)                // Get an entry by its key.
-	write(key string, values Values) error          // Write an entry to the store.
-	add(key string, entry *entry)                   // Add a new entry to the store.
-	remove(key string)                              // Remove an entry from the store.
-	keys(sorted bool) []string                      // Return an optionally sorted slice of entry keys.
-	apply(f func(string, *entry) error) error       // Apply f to all entries in the store in parallel.
-	applySerial(f func(string, *entry) error) error // Apply f to all entries in serial.
+	entry(key []byte) *entry                        // Get an entry by its key.
+	write(key []byte, values Values) (bool, error)  // Write an entry to the store.
+	add(key []byte, entry *entry)                   // Add a new entry to the store.
+	remove(key []byte)                              // Remove an entry from the store.
+	keys(sorted bool) [][]byte                      // Return an optionally sorted slice of entry keys.
+	apply(f func([]byte, *entry) error) error       // Apply f to all entries in the store in parallel.
+	applySerial(f func([]byte, *entry) error) error // Apply f to all entries in serial.
 	reset()                                         // Reset the store to an initial unused state.
+	split(n int) []storer                           // Split splits the store into n stores
+	count() int                                     // Count returns the number of keys in the store
 }
 
 // Cache maintains an in-memory store of Values for a set of keys.
@@ -201,18 +214,23 @@ type Cache struct {
 
 	stats        *CacheStatistics
 	lastSnapshot time.Time
+
+	// A one time synchronization used to initial the cache with a store.  Since the store can allocate a
+	// a large amount memory across shards, we lazily create it.
+	initialize       atomic.Value
+	initializedCount uint32
 }
 
 // NewCache returns an instance of a cache which will use a maximum of maxSize bytes of memory.
 // Only used for engine caches, never for snapshots.
 func NewCache(maxSize uint64, path string) *Cache {
-	store, _ := newring(ringShards)
 	c := &Cache{
 		maxSize:      maxSize,
-		store:        store, // Max size for now..
+		store:        emptyStore{},
 		stats:        &CacheStatistics{},
 		lastSnapshot: time.Now(),
 	}
+	c.initialize.Store(&sync.Once{})
 	c.UpdateAge()
 	c.UpdateCompactTime(0)
 	c.updateCachedBytes(0)
@@ -253,9 +271,33 @@ func (c *Cache) Statistics(tags map[string]string) []models.Statistic {
 	}}
 }
 
+// init initializes the cache and allocates the underlying store.  Once initialized,
+// the store re-used until Freed.
+func (c *Cache) init() {
+	if !atomic.CompareAndSwapUint32(&c.initializedCount, 0, 1) {
+		return
+	}
+
+	c.mu.Lock()
+	c.store, _ = newring(ringShards)
+	c.mu.Unlock()
+}
+
+// Free releases the underlying store and memory held by the Cache.
+func (c *Cache) Free() {
+	if !atomic.CompareAndSwapUint32(&c.initializedCount, 1, 0) {
+		return
+	}
+
+	c.mu.Lock()
+	c.store = emptyStore{}
+	c.mu.Unlock()
+}
+
 // Write writes the set of values for the key to the cache. This function is goroutine-safe.
 // It returns an error if the cache will exceed its max size by adding the new values.
-func (c *Cache) Write(key string, values []Value) error {
+func (c *Cache) Write(key []byte, values []Value) error {
+	c.init()
 	addedSize := uint64(Values(values).Size())
 
 	// Enough room in the cache?
@@ -267,11 +309,15 @@ func (c *Cache) Write(key string, values []Value) error {
 		return ErrCacheMemorySizeLimitExceeded(n, limit)
 	}
 
-	if err := c.store.write(key, values); err != nil {
+	newKey, err := c.store.write(key, values)
+	if err != nil {
 		atomic.AddInt64(&c.stats.WriteErr, 1)
 		return err
 	}
 
+	if newKey {
+		addedSize += uint64(len(key))
+	}
 	// Update the cache size and the memory size stat.
 	c.increaseSize(addedSize)
 	c.updateMemSize(int64(addedSize))
@@ -286,6 +332,7 @@ func (c *Cache) Write(key string, values []Value) error {
 // values as possible.  If one key fails, the others can still succeed and an
 // error will be returned.
 func (c *Cache) WriteMulti(values map[string][]Value) error {
+	c.init()
 	var addedSize uint64
 	for _, v := range values {
 		addedSize += uint64(Values(v).Size())
@@ -307,12 +354,17 @@ func (c *Cache) WriteMulti(values map[string][]Value) error {
 	// We'll optimistially set size here, and then decrement it for write errors.
 	c.increaseSize(addedSize)
 	for k, v := range values {
-		if err := store.write(k, v); err != nil {
+		newKey, err := store.write([]byte(k), v)
+		if err != nil {
 			// The write failed, hold onto the error and adjust the size delta.
 			werr = err
 			addedSize -= uint64(Values(v).Size())
 			c.decreaseSize(uint64(Values(v).Size()))
 		}
+		if newKey {
+			addedSize += uint64(len(k))
+			c.increaseSize(uint64(len(k)))
+		}
 	}
 
 	// Some points in the batch were dropped.  An error is returned so
@@ -332,6 +384,8 @@ func (c *Cache) WriteMulti(values map[string][]Value) error {
 // Snapshot takes a snapshot of the current cache, adds it to the slice of caches that
 // are being flushed, and resets the current cache with new values.
 func (c *Cache) Snapshot() (*Cache, error) {
+	c.init()
+
 	c.mu.Lock()
 	defer c.mu.Unlock()
 
@@ -388,12 +442,23 @@ func (c *Cache) Deduplicate() {
 
 	// Apply a function that simply calls deduplicate on each entry in the ring.
 	// apply cannot return an error in this invocation.
-	_ = store.apply(func(_ string, e *entry) error { e.deduplicate(); return nil })
+	_ = store.apply(func(_ []byte, e *entry) error { e.deduplicate(); return nil })
 }
 
 // ClearSnapshot removes the snapshot cache from the list of flushing caches and
 // adjusts the size.
 func (c *Cache) ClearSnapshot(success bool) {
+	c.init()
+
+	c.mu.RLock()
+	snapStore := c.snapshot.store
+	c.mu.RUnlock()
+
+	// reset the snapshot store outside of the write lock
+	if success {
+		snapStore.reset()
+	}
+
 	c.mu.Lock()
 	defer c.mu.Unlock()
 
@@ -403,8 +468,7 @@ func (c *Cache) ClearSnapshot(success bool) {
 		c.snapshotAttempts = 0
 		c.updateMemSize(-int64(atomic.LoadUint64(&c.snapshotSize))) // decrement the number of bytes in cache
 
-		// Reset the snapshot's store, and reset the snapshot to a fresh Cache.
-		c.snapshot.store.reset()
+		// Reset the snapshot to a fresh Cache.
 		c.snapshot = &Cache{
 			store: c.snapshot.store,
 		}
@@ -435,17 +499,39 @@ func (c *Cache) MaxSize() uint64 {
 	return c.maxSize
 }
 
+func (c *Cache) Count() int {
+	c.mu.RLock()
+	n := c.store.count()
+	c.mu.RUnlock()
+	return n
+}
+
 // Keys returns a sorted slice of all keys under management by the cache.
-func (c *Cache) Keys() []string {
+func (c *Cache) Keys() [][]byte {
 	c.mu.RLock()
 	store := c.store
 	c.mu.RUnlock()
 	return store.keys(true)
 }
 
+func (c *Cache) Split(n int) []*Cache {
+	if n == 1 {
+		return []*Cache{c}
+	}
+
+	caches := make([]*Cache, n)
+	storers := c.store.split(n)
+	for i := 0; i < n; i++ {
+		caches[i] = &Cache{
+			store: storers[i],
+		}
+	}
+	return caches
+}
+
 // unsortedKeys returns a slice of all keys under management by the cache. The
 // keys are not sorted.
-func (c *Cache) unsortedKeys() []string {
+func (c *Cache) unsortedKeys() [][]byte {
 	c.mu.RLock()
 	store := c.store
 	c.mu.RUnlock()
@@ -453,17 +539,17 @@ func (c *Cache) unsortedKeys() []string {
 }
 
 // Values returns a copy of all values, deduped and sorted, for the given key.
-func (c *Cache) Values(key string) Values {
+func (c *Cache) Values(key []byte) Values {
 	var snapshotEntries *entry
 
 	c.mu.RLock()
-	e, ok := c.store.entry(key)
+	e := c.store.entry(key)
 	if c.snapshot != nil {
-		snapshotEntries, _ = c.snapshot.store.entry(key)
+		snapshotEntries = c.snapshot.store.entry(key)
 	}
 	c.mu.RUnlock()
 
-	if !ok {
+	if e == nil {
 		if snapshotEntries == nil {
 			// No values in hot cache or snapshots.
 			return nil
@@ -510,7 +596,7 @@ func (c *Cache) Values(key string) Values {
 }
 
 // Delete removes all values for the given keys from the cache.
-func (c *Cache) Delete(keys []string) {
+func (c *Cache) Delete(keys [][]byte) {
 	c.DeleteRange(keys, math.MinInt64, math.MaxInt64)
 }
 
@@ -518,20 +604,22 @@ func (c *Cache) Delete(keys []string) {
 // with timestamps between between min and max from the cache.
 //
 // TODO(edd): Lock usage could possibly be optimised if necessary.
-func (c *Cache) DeleteRange(keys []string, min, max int64) {
+func (c *Cache) DeleteRange(keys [][]byte, min, max int64) {
+	c.init()
+
 	c.mu.Lock()
 	defer c.mu.Unlock()
 
 	for _, k := range keys {
 		// Make sure key exist in the cache, skip if it does not
-		e, ok := c.store.entry(k)
-		if !ok {
+		e := c.store.entry(k)
+		if e == nil {
 			continue
 		}
 
 		origSize := uint64(e.size())
 		if min == math.MinInt64 && max == math.MaxInt64 {
-			c.decreaseSize(origSize)
+			c.decreaseSize(origSize + uint64(len(k)))
 			c.store.remove(k)
 			continue
 		}
@@ -539,7 +627,7 @@ func (c *Cache) DeleteRange(keys []string, min, max int64) {
 		e.filter(min, max)
 		if e.count() == 0 {
 			c.store.remove(k)
-			c.decreaseSize(origSize)
+			c.decreaseSize(origSize + uint64(len(k)))
 			continue
 		}
 
@@ -558,8 +646,8 @@ func (c *Cache) SetMaxSize(size uint64) {
 // values returns the values for the key. It assumes the data is already sorted.
 // It doesn't lock the cache but it does read-lock the entry if there is one for the key.
 // values should only be used in compact.go in the CacheKeyIterator.
-func (c *Cache) values(key string) Values {
-	e, _ := c.store.entry(key)
+func (c *Cache) values(key []byte) Values {
+	e := c.store.entry(key)
 	if e == nil {
 		return nil
 	}
@@ -572,7 +660,7 @@ func (c *Cache) values(key string) Values {
 // ApplyEntryFn applies the function f to each entry in the Cache.
 // ApplyEntryFn calls f on each entry in turn, within the same goroutine.
 // It is safe for use by multiple goroutines.
-func (c *Cache) ApplyEntryFn(f func(key string, entry *entry) error) error {
+func (c *Cache) ApplyEntryFn(f func(key []byte, entry *entry) error) error {
 	c.mu.RLock()
 	store := c.store
 	c.mu.RUnlock()
@@ -601,12 +689,15 @@ func NewCacheLoader(files []string) *CacheLoader {
 // file is truncated up to and including the last valid byte, and processing
 // continues with the next segment file.
 func (cl *CacheLoader) Load(cache *Cache) error {
+
+	var r *WALSegmentReader
 	for _, fn := range cl.files {
 		if err := func() error {
 			f, err := os.OpenFile(fn, os.O_CREATE|os.O_RDWR, 0666)
 			if err != nil {
 				return err
 			}
+			defer f.Close()
 
 			// Log some information about the segments.
 			stat, err := os.Stat(f.Name())
@@ -615,8 +706,17 @@ func (cl *CacheLoader) Load(cache *Cache) error {
 			}
 			cl.Logger.Info(fmt.Sprintf("reading file %s, size %d", f.Name(), stat.Size()))
 
-			r := NewWALSegmentReader(f)
-			defer r.Close()
+			// Nothing to read, skip it
+			if stat.Size() == 0 {
+				return nil
+			}
+
+			if r == nil {
+				r = NewWALSegmentReader(f)
+				defer r.Close()
+			} else {
+				r.Reset(f)
+			}
 
 			for r.Next() {
 				entry, err := r.Read()
@@ -641,7 +741,7 @@ func (cl *CacheLoader) Load(cache *Cache) error {
 				}
 			}
 
-			return nil
+			return r.Close()
 		}(); err != nil {
 			return err
 		}
@@ -698,3 +798,16 @@ func (c *Cache) updateSnapshots() {
 	atomic.StoreInt64(&c.stats.DiskSizeBytes, int64(atomic.LoadUint64(&c.snapshotSize)))
 	atomic.StoreInt64(&c.stats.SnapshotCount, int64(c.snapshotAttempts))
 }
+
+type emptyStore struct{}
+
+func (e emptyStore) entry(key []byte) *entry                        { return nil }
+func (e emptyStore) write(key []byte, values Values) (bool, error)  { return false, nil }
+func (e emptyStore) add(key []byte, entry *entry)                   {}
+func (e emptyStore) remove(key []byte)                              {}
+func (e emptyStore) keys(sorted bool) [][]byte                      { return nil }
+func (e emptyStore) apply(f func([]byte, *entry) error) error       { return nil }
+func (e emptyStore) applySerial(f func([]byte, *entry) error) error { return nil }
+func (e emptyStore) reset()                                         {}
+func (e emptyStore) split(n int) []storer                           { return nil }
+func (e emptyStore) count() int                                     { return 0 }
