@@ -6,9 +6,9 @@ import (
 	"runtime"
 	"time"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/pkg/pool"
 	"github.com/influxdata/influxdb/tsdb"
+	"github.com/influxdata/influxql"
 )
 
 const (
@@ -24,6 +24,9 @@ const (
 	// BlockString designates a block encodes string values.
 	BlockString = byte(3)
 
+	// BlockUnsigned designates a block encodes uint64 values.
+	BlockUnsigned = byte(4)
+
 	// encodedBlockHeaderSize is the size of the header for an encoded block.  There is one
 	// byte encoding the type of the block.
 	encodedBlockHeaderSize = 1
@@ -114,6 +117,8 @@ func NewValue(t int64, value interface{}) Value {
 	switch v := value.(type) {
 	case int64:
 		return IntegerValue{unixnano: t, value: v}
+	case uint64:
+		return UnsignedValue{unixnano: t, value: v}
 	case float64:
 		return FloatValue{unixnano: t, value: v}
 	case bool:
@@ -129,6 +134,11 @@ func NewIntegerValue(t int64, v int64) Value {
 	return IntegerValue{unixnano: t, value: v}
 }
 
+// NewUnsignedValue returns a new unsigned integer value.
+func NewUnsignedValue(t int64, v uint64) Value {
+	return UnsignedValue{unixnano: t, value: v}
+}
+
 // NewFloatValue returns a new float value.
 func NewFloatValue(t int64, v float64) Value {
 	return FloatValue{unixnano: t, value: v}
@@ -159,11 +169,12 @@ func (e EmptyValue) Size() int { return 0 }
 // String returns the empty string.
 func (e EmptyValue) String() string { return "" }
 
-func (_ EmptyValue) internalOnly()   {}
-func (_ StringValue) internalOnly()  {}
-func (_ IntegerValue) internalOnly() {}
-func (_ BooleanValue) internalOnly() {}
-func (_ FloatValue) internalOnly()   {}
+func (_ EmptyValue) internalOnly()    {}
+func (_ StringValue) internalOnly()   {}
+func (_ IntegerValue) internalOnly()  {}
+func (_ UnsignedValue) internalOnly() {}
+func (_ BooleanValue) internalOnly()  {}
+func (_ FloatValue) internalOnly()    {}
 
 // Encode converts the values to a byte slice.  If there are no values,
 // this function panics.
@@ -177,6 +188,8 @@ func (a Values) Encode(buf []byte) ([]byte, error) {
 		return encodeFloatBlock(buf, a)
 	case IntegerValue:
 		return encodeIntegerBlock(buf, a)
+	case UnsignedValue:
+		return encodeUnsignedBlock(buf, a)
 	case BooleanValue:
 		return encodeBooleanBlock(buf, a)
 	case StringValue:
@@ -197,6 +210,8 @@ func (a Values) InfluxQLType() (influxql.DataType, error) {
 		return influxql.Float, nil
 	case IntegerValue:
 		return influxql.Integer, nil
+	case UnsignedValue:
+		return influxql.Unsigned, nil
 	case BooleanValue:
 		return influxql.Boolean, nil
 	case StringValue:
@@ -211,7 +226,7 @@ func (a Values) InfluxQLType() (influxql.DataType, error) {
 func BlockType(block []byte) (byte, error) {
 	blockType := block[0]
 	switch blockType {
-	case BlockFloat64, BlockInteger, BlockBoolean, BlockString:
+	case BlockFloat64, BlockInteger, BlockUnsigned, BlockBoolean, BlockString:
 		return blockType, nil
 	default:
 		return 0, fmt.Errorf("unknown block type: %d", blockType)
@@ -265,6 +280,17 @@ func DecodeBlock(block []byte, vals []Value) ([]Value, error) {
 		}
 		return vals[:len(decoded)], err
 
+	case BlockUnsigned:
+		var buf []UnsignedValue
+		decoded, err := DecodeUnsignedBlock(block, &buf)
+		if len(vals) < len(decoded) {
+			vals = make([]Value, len(decoded))
+		}
+		for i := range decoded {
+			vals[i] = decoded[i]
+		}
+		return vals[:len(decoded)], err
+
 	case BlockBoolean:
 		var buf []BooleanValue
 		decoded, err := DecodeBooleanBlock(block, &buf)
@@ -333,32 +359,7 @@ func encodeFloatBlock(buf []byte, values []Value) ([]byte, error) {
 	// frame-or-reference and run length encoding.
 	tsenc := getTimeEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			vv := v.(FloatValue)
-			tsenc.Write(vv.unixnano)
-			venc.Write(vv.value)
-		}
-		venc.Flush()
-
-		// Encoded timestamp values
-		tb, err := tsenc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded float values
-		vb, err := venc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes and the block
-		// in the next byte, followed by the block
-		b = packBlock(buf, BlockFloat64, tb, vb)
-
-		return nil
-	}()
+	b, err := encodeFloatBlockUsing(buf, values, tsenc, venc)
 
 	putTimeEncoder(tsenc)
 	putFloatEncoder(venc)
@@ -366,6 +367,33 @@ func encodeFloatBlock(buf []byte, values []Value) ([]byte, error) {
 	return b, err
 }
 
+func encodeFloatBlockUsing(buf []byte, values []Value, tsenc TimeEncoder, venc *FloatEncoder) ([]byte, error) {
+	tsenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(FloatValue)
+		tsenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+	venc.Flush()
+
+	// Encoded timestamp values
+	tb, err := tsenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded float values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes and the block
+	// in the next byte, followed by the block
+	return packBlock(buf, BlockFloat64, tb, vb), nil
+}
+
 // DecodeFloatBlock decodes the float block from the byte slice
 // and appends the float values to a.
 func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
@@ -381,11 +409,19 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]FloatValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := floatDecoderPool.Get(0).(*FloatDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []FloatValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		err = vdec.SetBytes(vb)
@@ -394,18 +430,12 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 		}
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Values()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, FloatValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = FloatValue{unixnano: tdec.Read(), value: vdec.Values()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -419,7 +449,7 @@ func DecodeFloatBlock(block []byte, a *[]FloatValue) ([]FloatValue, error) {
 			return err
 		}
 		return nil
-	}()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	floatDecoderPool.Put(vdec)
@@ -465,30 +495,7 @@ func encodeBooleanBlock(buf []byte, values []Value) ([]byte, error) {
 	// Encode timestamps using an adaptive encoder
 	tsenc := getTimeEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			vv := v.(BooleanValue)
-			tsenc.Write(vv.unixnano)
-			venc.Write(vv.value)
-		}
-
-		// Encoded timestamp values
-		tb, err := tsenc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded float values
-		vb, err := venc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes and the block
-		// in the next byte, followed by the block
-		b = packBlock(buf, BlockBoolean, tb, vb)
-		return nil
-	}()
+	b, err := encodeBooleanBlockUsing(buf, values, tsenc, venc)
 
 	putTimeEncoder(tsenc)
 	putBooleanEncoder(venc)
@@ -496,6 +503,32 @@ func encodeBooleanBlock(buf []byte, values []Value) ([]byte, error) {
 	return b, err
 }
 
+func encodeBooleanBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc BooleanEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(BooleanValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded float values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes and the block
+	// in the next byte, followed by the block
+	return packBlock(buf, BlockBoolean, tb, vb), nil
+}
+
 // DecodeBooleanBlock decodes the boolean block from the byte slice
 // and appends the boolean values to a.
 func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error) {
@@ -511,28 +544,30 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]BooleanValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := booleanDecoderPool.Get(0).(*BooleanDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []BooleanValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		vdec.SetBytes(vb)
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, BooleanValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = BooleanValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -545,7 +580,7 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 			return err
 		}
 		return nil
-	}()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	booleanDecoderPool.Put(vdec)
@@ -553,7 +588,7 @@ func DecodeBooleanBlock(block []byte, a *[]BooleanValue) ([]BooleanValue, error)
 	return (*a)[:i], err
 }
 
-// FloatValue represents an int64 value.
+// IntegerValue represents an int64 value.
 type IntegerValue struct {
 	unixnano int64
 	value    int64
@@ -580,39 +615,42 @@ func (v IntegerValue) String() string {
 }
 
 func encodeIntegerBlock(buf []byte, values []Value) ([]byte, error) {
-	tsEnc := getTimeEncoder(len(values))
-	vEnc := getIntegerEncoder(len(values))
+	tenc := getTimeEncoder(len(values))
+	venc := getIntegerEncoder(len(values))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			vv := v.(IntegerValue)
-			tsEnc.Write(vv.unixnano)
-			vEnc.Write(vv.value)
-		}
+	b, err := encodeIntegerBlockUsing(buf, values, tenc, venc)
 
-		// Encoded timestamp values
-		tb, err := tsEnc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded int64 values
-		vb, err := vEnc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes
-		b = packBlock(buf, BlockInteger, tb, vb)
-		return nil
-	}()
-
-	putTimeEncoder(tsEnc)
-	putIntegerEncoder(vEnc)
+	putTimeEncoder(tenc)
+	putIntegerEncoder(venc)
 
 	return b, err
 }
 
+func encodeIntegerBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc IntegerEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(IntegerValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded int64 values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockInteger, tb, vb), nil
+}
+
 // DecodeIntegerBlock decodes the integer block from the byte slice
 // and appends the integer values to a.
 func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error) {
@@ -629,28 +667,30 @@ func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error)
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]IntegerValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := integerDecoderPool.Get(0).(*IntegerDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []IntegerValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		vdec.SetBytes(vb)
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, IntegerValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = IntegerValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -663,7 +703,130 @@ func DecodeIntegerBlock(block []byte, a *[]IntegerValue) ([]IntegerValue, error)
 			return err
 		}
 		return nil
-	}()
+	}(*a)
+
+	timeDecoderPool.Put(tdec)
+	integerDecoderPool.Put(vdec)
+
+	return (*a)[:i], err
+}
+
+// UnsignedValue represents an int64 value.
+type UnsignedValue struct {
+	unixnano int64
+	value    uint64
+}
+
+// Value returns the underlying int64 value.
+func (v UnsignedValue) Value() interface{} {
+	return v.value
+}
+
+// UnixNano returns the timestamp of the value.
+func (v UnsignedValue) UnixNano() int64 {
+	return v.unixnano
+}
+
+// Size returns the number of bytes necessary to represent the value and its timestamp.
+func (v UnsignedValue) Size() int {
+	return 16
+}
+
+// String returns the string representation of the value and its timestamp.
+func (v UnsignedValue) String() string {
+	return fmt.Sprintf("%v %v", time.Unix(0, v.unixnano), v.Value())
+}
+
+func encodeUnsignedBlock(buf []byte, values []Value) ([]byte, error) {
+	tenc := getTimeEncoder(len(values))
+	venc := getUnsignedEncoder(len(values))
+
+	b, err := encodeUnsignedBlockUsing(buf, values, tenc, venc)
+
+	putTimeEncoder(tenc)
+	putUnsignedEncoder(venc)
+
+	return b, err
+}
+
+func encodeUnsignedBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc IntegerEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(UnsignedValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(int64(vv.value))
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded int64 values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockUnsigned, tb, vb), nil
+}
+
+// DecodeUnsignedBlock decodes the unsigned integer block from the byte slice
+// and appends the unsigned integer values to a.
+func DecodeUnsignedBlock(block []byte, a *[]UnsignedValue) ([]UnsignedValue, error) {
+	blockType := block[0]
+	if blockType != BlockUnsigned {
+		return nil, fmt.Errorf("invalid block type: exp %d, got %d", BlockUnsigned, blockType)
+	}
+
+	block = block[1:]
+
+	// The first 8 bytes is the minimum timestamp of the block
+	tb, vb, err := unpackBlock(block)
+	if err != nil {
+		return nil, err
+	}
+
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]UnsignedValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
+	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
+	vdec := integerDecoderPool.Get(0).(*IntegerDecoder)
+
+	var i int
+	err = func(a []UnsignedValue) error {
+		// Setup our timestamp and value decoders
+		tdec.Init(tb)
+		vdec.SetBytes(vb)
+
+		// Decode both a timestamp and value
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = UnsignedValue{unixnano: tdec.Read(), value: uint64(vdec.Read())}
+			j++
+		}
+		i = j
+
+		// Did timestamp decoding have an error?
+		err = tdec.Error()
+		if err != nil {
+			return err
+		}
+		// Did int64 decoding have an error?
+		err = vdec.Error()
+		if err != nil {
+			return err
+		}
+		return nil
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	integerDecoderPool.Put(vdec)
@@ -698,40 +861,42 @@ func (v StringValue) String() string {
 }
 
 func encodeStringBlock(buf []byte, values []Value) ([]byte, error) {
-	tsEnc := getTimeEncoder(len(values))
-	vEnc := getStringEncoder(len(values) * len(values[0].(StringValue).value))
+	tenc := getTimeEncoder(len(values))
+	venc := getStringEncoder(len(values) * len(values[0].(StringValue).value))
 
-	var b []byte
-	err := func() error {
-		for _, v := range values {
-			vv := v.(StringValue)
-			tsEnc.Write(vv.unixnano)
-			vEnc.Write(vv.value)
-		}
+	b, err := encodeStringBlockUsing(buf, values, tenc, venc)
 
-		// Encoded timestamp values
-		tb, err := tsEnc.Bytes()
-		if err != nil {
-			return err
-		}
-		// Encoded string values
-		vb, err := vEnc.Bytes()
-		if err != nil {
-			return err
-		}
-
-		// Prepend the first timestamp of the block in the first 8 bytes
-		b = packBlock(buf, BlockString, tb, vb)
-
-		return nil
-	}()
-
-	putTimeEncoder(tsEnc)
-	putStringEncoder(vEnc)
+	putTimeEncoder(tenc)
+	putStringEncoder(venc)
 
 	return b, err
 }
 
+func encodeStringBlockUsing(buf []byte, values []Value, tenc TimeEncoder, venc StringEncoder) ([]byte, error) {
+	tenc.Reset()
+	venc.Reset()
+
+	for _, v := range values {
+		vv := v.(StringValue)
+		tenc.Write(vv.unixnano)
+		venc.Write(vv.value)
+	}
+
+	// Encoded timestamp values
+	tb, err := tenc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+	// Encoded string values
+	vb, err := venc.Bytes()
+	if err != nil {
+		return nil, err
+	}
+
+	// Prepend the first timestamp of the block in the first 8 bytes
+	return packBlock(buf, BlockString, tb, vb), nil
+}
+
 // DecodeStringBlock decodes the string block from the byte slice
 // and appends the string values to a.
 func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
@@ -748,11 +913,19 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 		return nil, err
 	}
 
+	sz := CountTimestamps(tb)
+
+	if cap(*a) < sz {
+		*a = make([]StringValue, sz)
+	} else {
+		*a = (*a)[:sz]
+	}
+
 	tdec := timeDecoderPool.Get(0).(*TimeDecoder)
 	vdec := stringDecoderPool.Get(0).(*StringDecoder)
 
 	var i int
-	err = func() error {
+	err = func(a []StringValue) error {
 		// Setup our timestamp and value decoders
 		tdec.Init(tb)
 		err = vdec.SetBytes(vb)
@@ -761,18 +934,12 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 		}
 
 		// Decode both a timestamp and value
-		for tdec.Next() && vdec.Next() {
-			ts := tdec.Read()
-			v := vdec.Read()
-			if i < len(*a) {
-				elem := &(*a)[i]
-				elem.unixnano = ts
-				elem.value = v
-			} else {
-				*a = append(*a, StringValue{ts, v})
-			}
-			i++
+		j := 0
+		for j < len(a) && tdec.Next() && vdec.Next() {
+			a[j] = StringValue{unixnano: tdec.Read(), value: vdec.Read()}
+			j++
 		}
+		i = j
 
 		// Did timestamp decoding have an error?
 		err = tdec.Error()
@@ -785,7 +952,7 @@ func DecodeStringBlock(block []byte, a *[]StringValue) ([]StringValue, error) {
 			return err
 		}
 		return nil
-	}()
+	}(*a)
 
 	timeDecoderPool.Put(tdec)
 	stringDecoderPool.Put(vdec)
@@ -858,6 +1025,13 @@ func getIntegerEncoder(sz int) IntegerEncoder {
 }
 func putIntegerEncoder(enc IntegerEncoder) { integerEncoderPool.Put(enc) }
 
+func getUnsignedEncoder(sz int) IntegerEncoder {
+	x := integerEncoderPool.Get(sz).(IntegerEncoder)
+	x.Reset()
+	return x
+}
+func putUnsignedEncoder(enc IntegerEncoder) { integerEncoderPool.Put(enc) }
+
 func getFloatEncoder(sz int) *FloatEncoder {
 	x := floatEncoderPool.Get(sz).(*FloatEncoder)
 	x.Reset()
