@@ -15,11 +15,11 @@ import (
 
 	"github.com/influxdata/influxdb/pkg/estimator/hll"
 
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/pkg/bloom"
 	"github.com/influxdata/influxdb/pkg/estimator"
 	"github.com/influxdata/influxdb/pkg/mmap"
+	"github.com/influxdata/influxql"
 )
 
 // Log errors.
@@ -189,17 +189,17 @@ func (f *LogFile) Release() { f.wg.Done() }
 
 // Stat returns size and last modification time of the file.
 func (f *LogFile) Stat() (int64, time.Time) {
-	f.mu.Lock()
+	f.mu.RLock()
 	size, modTime := f.size, f.modTime
-	f.mu.Unlock()
+	f.mu.RUnlock()
 	return size, modTime
 }
 
 // Size returns the size of the file, in bytes.
 func (f *LogFile) Size() int64 {
-	f.mu.Lock()
+	f.mu.RLock()
 	v := f.size
-	f.mu.Unlock()
+	f.mu.RUnlock()
 	return v
 }
 
@@ -429,16 +429,55 @@ func (f *LogFile) DeleteTagValue(name, key, value []byte) error {
 
 // AddSeriesList adds a list of series to the log file in bulk.
 func (f *LogFile) AddSeriesList(names [][]byte, tagsSlice []models.Tags) error {
+	// Determine total size of names, keys, values.
+	var n int
+	for i := range names {
+		n += len(names[i])
+
+		tags := tagsSlice[i]
+		for j := range tags {
+			n += len(tags[j].Key) + len(tags[j].Value)
+		}
+	}
+
+	// Allocate names, keys, & values in one block.
+	buf := make([]byte, n)
+
+	// Clone all entries.
+	entries := make([]LogEntry, len(names))
+	for i := range names {
+		copy(buf, names[i])
+		clonedName := buf[:len(names[i])]
+		buf = buf[len(names[i]):]
+
+		// Clone tag set.
+		var clonedTags models.Tags
+		if len(tagsSlice[i]) > 0 {
+			clonedTags = make(models.Tags, len(tagsSlice[i]))
+			for j, tags := range tagsSlice[i] {
+				copy(buf, tags.Key)
+				key := buf[:len(tags.Key)]
+				buf = buf[len(tags.Key):]
+
+				copy(buf, tags.Value)
+				value := buf[:len(tags.Value)]
+				buf = buf[len(tags.Value):]
+
+				clonedTags[j] = models.Tag{Key: key, Value: value}
+			}
+		}
+
+		entries[i] = LogEntry{Name: clonedName, Tags: clonedTags}
+	}
+
 	f.mu.Lock()
 	defer f.mu.Unlock()
 
-	for i := range names {
-		// The name and tags are clone to prevent a memory leak
-		e := LogEntry{Name: []byte(string(names[i])), Tags: tagsSlice[i].Clone()}
-		if err := f.appendEntry(&e); err != nil {
+	for i := range entries {
+		if err := f.appendEntry(&entries[i]); err != nil {
 			return err
 		}
-		f.execEntry(&e)
+		f.execEntry(&entries[i])
 	}
 	return nil
 }
@@ -927,8 +966,16 @@ func (f *LogFile) writeTagsetTo(w io.Writer, name string, info *logFileCompactIn
 		tagSetInfo := mmInfo.tagSet[k]
 		assert(tagSetInfo != nil, "tag set info not found")
 
+		// Sort tag values.
+		values := make([]string, 0, len(tag.tagValues))
+		for v := range tag.tagValues {
+			values = append(values, v)
+		}
+		sort.Strings(values)
+
 		// Add each value.
-		for v, value := range tag.tagValues {
+		for _, v := range values {
+			value := tag.tagValues[v]
 			tagValueInfo := tagSetInfo.tagValues[v]
 			sort.Sort(uint32Slice(tagValueInfo.seriesIDs))
 
