@@ -2,12 +2,14 @@ package tsdb
 
 import (
 	"bytes"
+	"context"
 	"errors"
 	"fmt"
 	"io"
 	"math"
 	"path/filepath"
 	"regexp"
+	"runtime"
 	"sort"
 	"strings"
 	"sync"
@@ -15,10 +17,12 @@ import (
 	"time"
 
 	"github.com/gogo/protobuf/proto"
-	"github.com/influxdata/influxdb/influxql"
 	"github.com/influxdata/influxdb/models"
 	"github.com/influxdata/influxdb/pkg/estimator"
+	"github.com/influxdata/influxdb/pkg/limiter"
+	"github.com/influxdata/influxdb/query"
 	internal "github.com/influxdata/influxdb/tsdb/internal"
+	"github.com/influxdata/influxql"
 	"github.com/uber-go/zap"
 )
 
@@ -109,16 +113,15 @@ type Shard struct {
 	path    string
 	walPath string
 	id      uint64
-	wg      sync.WaitGroup
 
 	database        string
 	retentionPolicy string
 
 	options EngineOptions
 
-	mu     sync.RWMutex
-	engine Engine
-	index  Index
+	mu      sync.RWMutex
+	_engine Engine
+	index   Index
 
 	closing chan struct{}
 	enabled bool
@@ -168,8 +171,9 @@ func NewShard(id uint64, path string, walPath string, opt EngineOptions) *Shard
 // WithLogger sets the logger on the shard.
 func (s *Shard) WithLogger(log zap.Logger) {
 	s.baseLogger = log
-	if err := s.ready(); err == nil {
-		s.engine.WithLogger(s.baseLogger)
+	engine, err := s.engine()
+	if err == nil {
+		engine.WithLogger(s.baseLogger)
 		s.index.WithLogger(s.baseLogger)
 	}
 	s.logger = s.baseLogger.With(zap.String("service", "shard"))
@@ -181,9 +185,9 @@ func (s *Shard) SetEnabled(enabled bool) {
 	s.mu.Lock()
 	// Prevent writes and queries
 	s.enabled = enabled
-	if s.engine != nil {
+	if s._engine != nil {
 		// Disable background compactions and snapshotting
-		s.engine.SetEnabled(enabled)
+		s._engine.SetEnabled(enabled)
 	}
 	s.mu.Unlock()
 }
@@ -218,13 +222,16 @@ type ShardStatistics struct {
 
 // Statistics returns statistics for periodic monitoring.
 func (s *Shard) Statistics(tags map[string]string) []models.Statistic {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return nil
 	}
 
 	// Refresh our disk size stat
-	_, _ = s.DiskSize()
-	seriesN := s.engine.SeriesN()
+	if _, err := s.DiskSize(); err != nil {
+		return nil
+	}
+	seriesN := engine.SeriesN()
 
 	tags = s.defaultTags.Merge(tags)
 	statistics := []models.Statistic{{
@@ -245,7 +252,7 @@ func (s *Shard) Statistics(tags map[string]string) []models.Statistic {
 	}}
 
 	// Add the index and engine statistics.
-	statistics = append(statistics, s.engine.Statistics(tags)...)
+	statistics = append(statistics, engine.Statistics(tags)...)
 	return statistics
 }
 
@@ -259,7 +266,7 @@ func (s *Shard) Open() error {
 		defer s.mu.Unlock()
 
 		// Return if the shard is already open
-		if s.engine != nil {
+		if s._engine != nil {
 			return nil
 		}
 
@@ -298,7 +305,7 @@ func (s *Shard) Open() error {
 		if err := e.LoadMetadataIndex(s.id, s.index); err != nil {
 			return err
 		}
-		s.engine = e
+		s._engine = e
 
 		return nil
 	}(); err != nil {
@@ -335,7 +342,7 @@ func (s *Shard) CloseFast() error {
 // close closes the shard an removes reference to the shard from associated
 // indexes, unless clean is false.
 func (s *Shard) close(clean bool) error {
-	if s.engine == nil {
+	if s._engine == nil {
 		return nil
 	}
 
@@ -345,16 +352,15 @@ func (s *Shard) close(clean bool) error {
 	default:
 		close(s.closing)
 	}
-	s.wg.Wait()
 
 	if clean {
 		// Don't leak our shard ID and series keys in the index
-		s.UnloadIndex()
+		s.index.RemoveShard(s.id)
 	}
 
-	err := s.engine.Close()
+	err := s._engine.Close()
 	if err == nil {
-		s.engine = nil
+		s._engine = nil
 	}
 
 	if e := s.index.Close(); e == nil {
@@ -364,12 +370,12 @@ func (s *Shard) close(clean bool) error {
 }
 
 func (s *Shard) IndexType() string {
+	s.mu.RLock()
+	defer s.mu.RUnlock()
 	if err := s.ready(); err != nil {
 		return ""
 	}
 
-	s.mu.RLock()
-	defer s.mu.RUnlock()
 	return s.index.Type()
 }
 
@@ -377,50 +383,81 @@ func (s *Shard) IndexType() string {
 // It returns nil if ready, otherwise ErrShardClosed or ErrShardDiabled
 func (s *Shard) ready() error {
 	var err error
-
-	s.mu.RLock()
-	if s.engine == nil {
+	if s._engine == nil {
 		err = ErrEngineClosed
 	} else if !s.enabled {
 		err = ErrShardDisabled
 	}
-	s.mu.RUnlock()
 	return err
 }
 
 // LastModified returns the time when this shard was last modified.
 func (s *Shard) LastModified() time.Time {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return time.Time{}
 	}
-	return s.engine.LastModified()
+	return engine.LastModified()
 }
 
 // UnloadIndex removes all references to this shard from the DatabaseIndex
 func (s *Shard) UnloadIndex() {
+	s.mu.RLock()
+	defer s.mu.RUnlock()
+	if err := s.ready(); err != nil {
+		return
+	}
 	s.index.RemoveShard(s.id)
 }
 
+// Index returns a reference to the underlying index.
+// This should only be used by utilities and not directly accessed by the database.
+func (s *Shard) Index() Index {
+	s.mu.RLock()
+	defer s.mu.RUnlock()
+	return s.index
+}
+
 // IsIdle return true if the shard is not receiving writes and is fully compacted.
 func (s *Shard) IsIdle() bool {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return true
 	}
+	return engine.IsIdle()
+}
 
-	return s.engine.IsIdle()
+func (s *Shard) Free() error {
+	engine, err := s.engine()
+	if err != nil {
+		return err
+	}
+
+	// Disable compactions to stop background goroutines
+	s.SetCompactionsEnabled(false)
+
+	return engine.Free()
 }
 
 // SetCompactionsEnabled enables or disable shard background compactions.
 func (s *Shard) SetCompactionsEnabled(enabled bool) {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return
 	}
-	s.engine.SetCompactionsEnabled(enabled)
+	engine.SetCompactionsEnabled(enabled)
 }
 
-// DiskSize returns the size on disk of this shard
+// DiskSize returns the size on disk of this shard.
 func (s *Shard) DiskSize() (int64, error) {
-	size := s.engine.DiskSize()
+	s.mu.RLock()
+	defer s.mu.RUnlock()
+	// We don't use engine() becuase we still want to report the shard's disk
+	// size even if the shard has been disabled.
+	if s._engine == nil {
+		return 0, ErrEngineClosed
+	}
+	size := s._engine.DiskSize()
 	atomic.StoreInt64(&s.stats.DiskBytes, size)
 	return size, nil
 }
@@ -433,15 +470,15 @@ type FieldCreate struct {
 
 // WritePoints will write the raw data points and any new metadata to the index in the shard.
 func (s *Shard) WritePoints(points []models.Point) error {
-	if err := s.ready(); err != nil {
+	s.mu.RLock()
+	defer s.mu.RUnlock()
+
+	engine, err := s.engineNoLock()
+	if err != nil {
 		return err
 	}
 
 	var writeError error
-
-	s.mu.RLock()
-	defer s.mu.RUnlock()
-
 	atomic.AddInt64(&s.stats.WriteReq, 1)
 
 	points, fieldsToCreate, err := s.validateSeriesAndFields(points)
@@ -461,7 +498,7 @@ func (s *Shard) WritePoints(points []models.Point) error {
 	}
 
 	// Write to the engine.
-	if err := s.engine.WritePoints(points); err != nil {
+	if err := engine.WritePoints(points); err != nil {
 		atomic.AddInt64(&s.stats.WritePointsErr, int64(len(points)))
 		atomic.AddInt64(&s.stats.WriteReqErr, 1)
 		return fmt.Errorf("engine: %s", err)
@@ -472,75 +509,6 @@ func (s *Shard) WritePoints(points []models.Point) error {
 	return writeError
 }
 
-// DeleteSeries deletes a list of series.
-func (s *Shard) DeleteSeries(seriesKeys [][]byte) error {
-	return s.DeleteSeriesRange(seriesKeys, math.MinInt64, math.MaxInt64)
-}
-
-// DeleteSeriesRange deletes all values from for seriesKeys between min and max (inclusive)
-func (s *Shard) DeleteSeriesRange(seriesKeys [][]byte, min, max int64) error {
-	if err := s.ready(); err != nil {
-		return err
-	}
-
-	if err := s.engine.DeleteSeriesRange(seriesKeys, min, max); err != nil {
-		return err
-	}
-
-	return nil
-}
-
-// DeleteMeasurement deletes a measurement and all underlying series.
-func (s *Shard) DeleteMeasurement(name []byte) error {
-	if err := s.ready(); err != nil {
-		return err
-	}
-	return s.engine.DeleteMeasurement(name)
-}
-
-// SeriesN returns the unique number of series in the shard.
-func (s *Shard) SeriesN() int64 {
-	return s.engine.SeriesN()
-}
-
-// SeriesSketches returns the series sketches for the shard.
-func (s *Shard) SeriesSketches() (estimator.Sketch, estimator.Sketch, error) {
-	s.mu.RLock()
-	defer s.mu.RUnlock()
-	if s.engine == nil {
-		return nil, nil, nil
-	}
-	return s.engine.SeriesSketches()
-}
-
-// MeasurementsSketches returns the measurement sketches for the shard.
-func (s *Shard) MeasurementsSketches() (estimator.Sketch, estimator.Sketch, error) {
-	s.mu.RLock()
-	defer s.mu.RUnlock()
-	if s.engine == nil {
-		return nil, nil, nil
-	}
-	return s.engine.MeasurementsSketches()
-}
-
-func (s *Shard) createFieldsAndMeasurements(fieldsToCreate []*FieldCreate) error {
-	if len(fieldsToCreate) == 0 {
-		return nil
-	}
-
-	// add fields
-	for _, f := range fieldsToCreate {
-		mf := s.engine.MeasurementFields(f.Measurement)
-		if err := mf.CreateFieldIfNotExists([]byte(f.Field.Name), f.Field.Type, false); err != nil {
-			return err
-		}
-
-		s.index.SetFieldName(f.Measurement, f.Field.Name)
-	}
-
-	return nil
-}
-
 // validateSeriesAndFields checks which series and fields are new and whose metadata should be saved and indexed.
 func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point, []*FieldCreate, error) {
 	var (
@@ -574,9 +542,14 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 	}
 	points, keys, names, tagsSlice = points[:j], keys[:j], names[:j], tagsSlice[:j]
 
+	engine, err := s.engineNoLock()
+	if err != nil {
+		return nil, nil, err
+	}
+
 	// Add new series. Check for partial writes.
 	var droppedKeys map[string]struct{}
-	if err := s.engine.CreateSeriesListIfNotExists(keys, names, tagsSlice); err != nil {
+	if err := engine.CreateSeriesListIfNotExists(keys, names, tagsSlice); err != nil {
 		switch err := err.(type) {
 		case *PartialWriteError:
 			reason = err.Reason
@@ -628,7 +601,7 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 		// see if the field definitions need to be saved to the shard
 		mf := mfCache[string(name)]
 		if mf == nil {
-			mf = s.engine.MeasurementFields(name).Clone()
+			mf = engine.MeasurementFields(name).Clone()
 			mfCache[string(name)] = mf
 		}
 		iter.Reset()
@@ -647,6 +620,8 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 				fieldType = influxql.Float
 			case models.Integer:
 				fieldType = influxql.Integer
+			case models.Unsigned:
+				fieldType = influxql.Unsigned
 			case models.Boolean:
 				fieldType = influxql.Boolean
 			case models.String:
@@ -688,66 +663,204 @@ func (s *Shard) validateSeriesAndFields(points []models.Point) ([]models.Point,
 	return points, fieldsToCreate, err
 }
 
+func (s *Shard) createFieldsAndMeasurements(fieldsToCreate []*FieldCreate) error {
+	if len(fieldsToCreate) == 0 {
+		return nil
+	}
+
+	engine, err := s.engineNoLock()
+	if err != nil {
+		return err
+	}
+
+	// add fields
+	for _, f := range fieldsToCreate {
+		mf := engine.MeasurementFields(f.Measurement)
+		if err := mf.CreateFieldIfNotExists([]byte(f.Field.Name), f.Field.Type, false); err != nil {
+			return err
+		}
+
+		s.index.SetFieldName(f.Measurement, f.Field.Name)
+	}
+
+	return nil
+}
+
+// DeleteSeries deletes a list of series.
+func (s *Shard) DeleteSeries(seriesKeys [][]byte) error {
+	return s.DeleteSeriesRange(seriesKeys, math.MinInt64, math.MaxInt64)
+}
+
+// DeleteSeriesRange deletes all values from for seriesKeys between min and max (inclusive)
+func (s *Shard) DeleteSeriesRange(seriesKeys [][]byte, min, max int64) error {
+	engine, err := s.engine()
+	if err != nil {
+		return err
+	}
+	return engine.DeleteSeriesRange(seriesKeys, min, max)
+}
+
+// DeleteMeasurement deletes a measurement and all underlying series.
+func (s *Shard) DeleteMeasurement(name []byte) error {
+	engine, err := s.engine()
+	if err != nil {
+		return err
+	}
+	return engine.DeleteMeasurement(name)
+}
+
+// SeriesN returns the unique number of series in the shard.
+func (s *Shard) SeriesN() int64 {
+	engine, err := s.engine()
+	if err != nil {
+		return 0
+	}
+	return engine.SeriesN()
+}
+
+// SeriesSketches returns the series sketches for the shard.
+func (s *Shard) SeriesSketches() (estimator.Sketch, estimator.Sketch, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, nil, err
+	}
+	return engine.SeriesSketches()
+}
+
+// MeasurementsSketches returns the measurement sketches for the shard.
+func (s *Shard) MeasurementsSketches() (estimator.Sketch, estimator.Sketch, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, nil, err
+	}
+	return engine.MeasurementsSketches()
+}
+
 // MeasurementNamesByExpr returns names of measurements matching the condition.
 // If cond is nil then all measurement names are returned.
-func (s *Shard) MeasurementNamesByExpr(cond influxql.Expr) ([][]byte, error) {
-	return s.engine.MeasurementNamesByExpr(cond)
+func (s *Shard) MeasurementNamesByExpr(auth query.Authorizer, cond influxql.Expr) ([][]byte, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+	return engine.MeasurementNamesByExpr(auth, cond)
+}
+
+// MeasurementNamesByRegex returns names of measurements matching the regular expression.
+func (s *Shard) MeasurementNamesByRegex(re *regexp.Regexp) ([][]byte, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+	return engine.MeasurementNamesByRegex(re)
+}
+
+// MeasurementSeriesKeysByExpr returns a list of series keys from the shard
+// matching expr.
+func (s *Shard) MeasurementSeriesKeysByExpr(name []byte, expr influxql.Expr) ([][]byte, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+	return engine.MeasurementSeriesKeysByExpr(name, expr)
+}
+
+// TagKeyHasAuthorizedSeries determines if there exists an authorised series on
+// the provided measurement with the provided tag key.
+func (s *Shard) TagKeyHasAuthorizedSeries(auth query.Authorizer, name []byte, key string) bool {
+	engine, err := s.engine()
+	if err != nil {
+		return false
+	}
+	return engine.TagKeyHasAuthorizedSeries(auth, name, key)
+}
+
+// MeasurementTagKeysByExpr returns all the tag keys for the provided expression.
+func (s *Shard) MeasurementTagKeysByExpr(name []byte, expr influxql.Expr) (map[string]struct{}, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+	return engine.MeasurementTagKeysByExpr(name, expr)
+}
+
+// MeasurementTagKeyValuesByExpr returns all the tag keys values for the
+// provided expression.
+func (s *Shard) MeasurementTagKeyValuesByExpr(auth query.Authorizer, name []byte, key []string, expr influxql.Expr, keysSorted bool) ([][]string, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+	return engine.MeasurementTagKeyValuesByExpr(auth, name, key, expr, keysSorted)
 }
 
 // MeasurementFields returns fields for a measurement.
+// TODO(edd): This method is currently only being called from tests; do we
+// really need it?
 func (s *Shard) MeasurementFields(name []byte) *MeasurementFields {
-	return s.engine.MeasurementFields(name)
+	engine, err := s.engine()
+	if err != nil {
+		return nil
+	}
+	return engine.MeasurementFields(name)
 }
 
+// MeasurementExists returns true if the shard contains name.
+// TODO(edd): This method is currently only being called from tests; do we
+// really need it?
 func (s *Shard) MeasurementExists(name []byte) (bool, error) {
-	return s.engine.MeasurementExists(name)
+	engine, err := s.engine()
+	if err != nil {
+		return false, err
+	}
+	return engine.MeasurementExists(name)
 }
 
 // WriteTo writes the shard's data to w.
 func (s *Shard) WriteTo(w io.Writer) (int64, error) {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return 0, err
 	}
-	n, err := s.engine.WriteTo(w)
+	n, err := engine.WriteTo(w)
 	atomic.AddInt64(&s.stats.BytesWritten, int64(n))
 	return n, err
 }
 
 // CreateIterator returns an iterator for the data in the shard.
-func (s *Shard) CreateIterator(measurement string, opt influxql.IteratorOptions) (influxql.Iterator, error) {
-	if err := s.ready(); err != nil {
+func (s *Shard) CreateIterator(ctx context.Context, m *influxql.Measurement, opt query.IteratorOptions) (query.Iterator, error) {
+	engine, err := s.engine()
+	if err != nil {
 		return nil, err
 	}
 
-	if strings.HasPrefix(measurement, "_") {
-		if itr, ok, err := s.createSystemIterator(measurement, opt); ok {
-			return itr, err
-		}
-		// Unknown system source so pass this to the engine.
+	switch m.SystemIterator {
+	case "_fieldKeys":
+		return NewFieldKeysIterator(engine, opt)
+	case "_series":
+		return s.createSeriesIterator(opt)
+	case "_tagKeys":
+		return NewTagKeysIterator(engine, opt)
 	}
-	return s.engine.CreateIterator(measurement, opt)
+	return engine.CreateIterator(ctx, m.Name, opt)
 }
 
-// createSystemIterator returns an iterator for a system source.
-func (s *Shard) createSystemIterator(measurement string, opt influxql.IteratorOptions) (influxql.Iterator, bool, error) {
-	switch measurement {
-	case "_fieldKeys":
-		itr, err := NewFieldKeysIterator(s, opt)
-		return itr, true, err
-	case "_series":
-		itr, err := s.createSeriesIterator(opt)
-		return itr, true, err
-	case "_tagKeys":
-		itr, err := NewTagKeysIterator(s, opt)
-		return itr, true, err
+func (s *Shard) CreateCursor(ctx context.Context, r *CursorRequest) (Cursor, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
 	}
-	return nil, false, nil
+	return engine.CreateCursor(ctx, r)
 }
 
 // createSeriesIterator returns a new instance of SeriesIterator.
-func (s *Shard) createSeriesIterator(opt influxql.IteratorOptions) (influxql.Iterator, error) {
+func (s *Shard) createSeriesIterator(opt query.IteratorOptions) (query.Iterator, error) {
+	engine, err := s.engine()
+	if err != nil {
+		return nil, err
+	}
+
 	// Only equality operators are allowed.
-	var err error
 	influxql.WalkFunc(opt.Condition, func(n influxql.Node) {
 		switch n := n.(type) {
 		case *influxql.BinaryExpr:
@@ -763,12 +876,13 @@ func (s *Shard) createSeriesIterator(opt influxql.IteratorOptions) (influxql.Ite
 		return nil, err
 	}
 
-	return s.engine.SeriesPointIterator(opt)
+	return engine.SeriesPointIterator(opt)
 }
 
 // FieldDimensions returns unique sets of fields and dimensions across a list of sources.
 func (s *Shard) FieldDimensions(measurements []string) (fields map[string]influxql.DataType, dimensions map[string]struct{}, err error) {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return nil, nil, err
 	}
 
@@ -790,7 +904,7 @@ func (s *Shard) FieldDimensions(measurements []string) (fields map[string]influx
 
 			if len(keys) > 0 {
 				for _, k := range keys {
-					if _, ok := fields[k]; !ok || influxql.String < fields[k] {
+					if fields[k].LessThan(influxql.String) {
 						fields[k] = influxql.String
 					}
 				}
@@ -800,23 +914,23 @@ func (s *Shard) FieldDimensions(measurements []string) (fields map[string]influx
 		}
 
 		// Retrieve measurement.
-		if exists, err := s.engine.MeasurementExists([]byte(name)); err != nil {
+		if exists, err := engine.MeasurementExists([]byte(name)); err != nil {
 			return nil, nil, err
 		} else if !exists {
 			continue
 		}
 
 		// Append fields and dimensions.
-		mf := s.engine.MeasurementFields([]byte(name))
+		mf := engine.MeasurementFields([]byte(name))
 		if mf != nil {
 			for k, typ := range mf.FieldSet() {
-				if _, ok := fields[k]; !ok || typ < fields[k] {
+				if fields[k].LessThan(typ) {
 					fields[k] = typ
 				}
 			}
 		}
 
-		if err := s.engine.ForEachMeasurementTagKey([]byte(name), func(key []byte) error {
+		if err := engine.ForEachMeasurementTagKey([]byte(name), func(key []byte) error {
 			dimensions[string(key)] = struct{}{}
 			return nil
 		}); err != nil {
@@ -827,62 +941,65 @@ func (s *Shard) FieldDimensions(measurements []string) (fields map[string]influx
 	return fields, dimensions, nil
 }
 
-func (s *Shard) MeasurementsByRegex(re *regexp.Regexp) []string {
-	a, _ := s.engine.MeasurementNamesByRegex(re)
+// mapType returns the data type for the field within the measurement.
+func (s *Shard) mapType(measurement, field string) (influxql.DataType, error) {
+	engine, err := s.engineNoLock()
+	if err != nil {
+		return 0, err
+	}
 
-	other := make([]string, len(a))
-	for i := range a {
-		other[i] = string(a[i])
+	switch field {
+	case "_name", "_tagKey", "_tagValue", "_seriesKey":
+		return influxql.String, nil
 	}
-	return other
-}
 
-// MapType returns the data type for the field within the measurement.
-func (s *Shard) MapType(measurement, field string) influxql.DataType {
 	// Process system measurements.
-	if strings.HasPrefix(measurement, "_") {
-		switch measurement {
-		case "_fieldKeys":
-			if field == "fieldKey" || field == "fieldType" {
-				return influxql.String
-			}
-			return influxql.Unknown
-		case "_series":
-			if field == "key" {
-				return influxql.String
-			}
-			return influxql.Unknown
-		case "_tagKeys":
-			if field == "tagKey" {
-				return influxql.String
-			}
-			return influxql.Unknown
+	switch measurement {
+	case "_fieldKeys":
+		if field == "fieldKey" || field == "fieldType" {
+			return influxql.String, nil
 		}
-		// Unknown system source so default to looking for a measurement.
+		return influxql.Unknown, nil
+	case "_series":
+		if field == "key" {
+			return influxql.String, nil
+		}
+		return influxql.Unknown, nil
+	case "_tagKeys":
+		if field == "tagKey" {
+			return influxql.String, nil
+		}
+		return influxql.Unknown, nil
+	}
+	// Unknown system source so default to looking for a measurement.
+
+	if exists, _ := engine.MeasurementExists([]byte(measurement)); !exists {
+		return influxql.Unknown, nil
 	}
 
-	if exists, _ := s.engine.MeasurementExists([]byte(measurement)); !exists {
-		return influxql.Unknown
-	}
-
-	mf := s.engine.MeasurementFields([]byte(measurement))
+	mf := engine.MeasurementFields([]byte(measurement))
 	if mf != nil {
 		f := mf.Field(field)
 		if f != nil {
-			return f.Type
+			return f.Type, nil
 		}
 	}
 
-	if exists, _ := s.engine.HasTagKey([]byte(measurement), []byte(field)); exists {
-		return influxql.Tag
+	if exists, _ := engine.HasTagKey([]byte(measurement), []byte(field)); exists {
+		return influxql.Tag, nil
 	}
 
-	return influxql.Unknown
+	return influxql.Unknown, nil
 }
 
-// ExpandSources expands regex sources and removes duplicates.
+// expandSources expands regex sources and removes duplicates.
 // NOTE: sources must be normalized (db and rp set) before calling this function.
-func (s *Shard) ExpandSources(sources influxql.Sources) (influxql.Sources, error) {
+func (s *Shard) expandSources(sources influxql.Sources) (influxql.Sources, error) {
+	engine, err := s.engineNoLock()
+	if err != nil {
+		return nil, err
+	}
+
 	// Use a map as a set to prevent duplicates.
 	set := map[string]influxql.Source{}
 
@@ -897,7 +1014,7 @@ func (s *Shard) ExpandSources(sources influxql.Sources) (influxql.Sources, error
 			}
 
 			// Loop over matching measurements.
-			names, err := s.engine.MeasurementNamesByRegex(src.Regex.Val)
+			names, err := engine.MeasurementNamesByRegex(src.Regex.Val)
 			if err != nil {
 				return nil, err
 			}
@@ -932,19 +1049,36 @@ func (s *Shard) ExpandSources(sources influxql.Sources) (influxql.Sources, error
 	return expanded, nil
 }
 
+// Backup backs up the shard by creating a tar archive of all TSM files that
+// have been modified since the provided time. See Engine.Backup for more details.
+func (s *Shard) Backup(w io.Writer, basePath string, since time.Time) error {
+	engine, err := s.engine()
+	if err != nil {
+		return err
+	}
+	return engine.Backup(w, basePath, since)
+}
+
 // Restore restores data to the underlying engine for the shard.
 // The shard is reopened after restore.
 func (s *Shard) Restore(r io.Reader, basePath string) error {
-	s.mu.Lock()
+	if err := func() error {
+		s.mu.Lock()
+		defer s.mu.Unlock()
 
-	// Restore to engine.
-	if err := s.engine.Restore(r, basePath); err != nil {
-		s.mu.Unlock()
+		// Special case - we can still restore to a disabled shard, so we should
+		// only check if the engine is closed and not care if the shard is
+		// disabled.
+		if s._engine == nil {
+			return ErrEngineClosed
+		}
+
+		// Restore to engine.
+		return s._engine.Restore(r, basePath)
+	}(); err != nil {
 		return err
 	}
 
-	s.mu.Unlock()
-
 	// Close shard.
 	if err := s.Close(); err != nil {
 		return err
@@ -957,42 +1091,83 @@ func (s *Shard) Restore(r io.Reader, basePath string) error {
 // Import imports data to the underlying engine for the shard. r should
 // be a reader from a backup created by Backup.
 func (s *Shard) Import(r io.Reader, basePath string) error {
+	// Special case - we can still import to a disabled shard, so we should
+	// only check if the engine is closed and not care if the shard is
+	// disabled.
 	s.mu.Lock()
 	defer s.mu.Unlock()
+	if s._engine == nil {
+		return ErrEngineClosed
+	}
 
-	// Restore to engine.
-	return s.engine.Import(r, basePath)
+	// Import to engine.
+	return s._engine.Import(r, basePath)
 }
 
 // CreateSnapshot will return a path to a temp directory
 // containing hard links to the underlying shard files.
 func (s *Shard) CreateSnapshot() (string, error) {
-	s.mu.RLock()
-	defer s.mu.RUnlock()
-	return s.engine.CreateSnapshot()
+	engine, err := s.engine()
+	if err != nil {
+		return "", err
+	}
+	return engine.CreateSnapshot()
+}
+
+// ForEachMeasurementName iterates over each measurement in the shard.
+func (s *Shard) ForEachMeasurementName(fn func(name []byte) error) error {
+	engine, err := s.engine()
+	if err != nil {
+		return err
+	}
+	return engine.ForEachMeasurementName(fn)
 }
 
 func (s *Shard) ForEachMeasurementTagKey(name []byte, fn func(key []byte) error) error {
-	if err := s.ready(); err != nil {
-		return nil
+	engine, err := s.engine()
+	if err != nil {
+		return err
 	}
-
-	return s.engine.ForEachMeasurementTagKey(name, fn)
+	return engine.ForEachMeasurementTagKey(name, fn)
 }
 
 func (s *Shard) TagKeyCardinality(name, key []byte) int {
-	if err := s.ready(); err != nil {
+	engine, err := s.engine()
+	if err != nil {
 		return 0
 	}
+	return engine.TagKeyCardinality(name, key)
+}
 
-	return s.engine.TagKeyCardinality(name, key)
+// engine safely (under an RLock) returns a reference to the shard's Engine, or
+// an error if the Engine is closed, or the shard is currently disabled.
+//
+// The shard's Engine should always be accessed via a call to engine(), rather
+// than directly referencing Shard.engine.
+//
+// If a caller needs an Engine reference but is already under a lock, then they
+// should use engineNoLock().
+func (s *Shard) engine() (Engine, error) {
+	s.mu.RLock()
+	defer s.mu.RUnlock()
+	return s.engineNoLock()
+}
+
+// engineNoLock is similar to calling engine(), but the caller must guarantee
+// that they already hold an appropriate lock.
+func (s *Shard) engineNoLock() (Engine, error) {
+	if err := s.ready(); err != nil {
+		return nil, err
+	}
+	return s._engine, nil
 }
 
 type ShardGroup interface {
 	MeasurementsByRegex(re *regexp.Regexp) []string
 	FieldDimensions(measurements []string) (fields map[string]influxql.DataType, dimensions map[string]struct{}, err error)
 	MapType(measurement, field string) influxql.DataType
-	CreateIterator(measurement string, opt influxql.IteratorOptions) (influxql.Iterator, error)
+	CreateIterator(ctx context.Context, measurement *influxql.Measurement, opt query.IteratorOptions) (query.Iterator, error)
+	IteratorCost(measurement string, opt query.IteratorOptions) (query.IteratorCost, error)
 	ExpandSources(sources influxql.Sources) (influxql.Sources, error)
 }
 
@@ -1008,12 +1183,22 @@ func (a Shards) Less(i, j int) bool { return a[i].id < a[j].id }
 // Swap implements sort.Interface.
 func (a Shards) Swap(i, j int) { a[i], a[j] = a[j], a[i] }
 
+// MeasurementsByRegex returns the unique set of measurements matching the
+// provided regex, for all the shards.
 func (a Shards) MeasurementsByRegex(re *regexp.Regexp) []string {
-	m := make(map[string]struct{})
+	var m map[string]struct{}
 	for _, sh := range a {
-		names := sh.MeasurementsByRegex(re)
+		names, err := sh.MeasurementNamesByRegex(re)
+		if err != nil {
+			continue // Skip this shard's results—previous behaviour.
+		}
+
+		if m == nil {
+			m = make(map[string]struct{}, len(names))
+		}
+
 		for _, name := range names {
-			m[name] = struct{}{}
+			m[string(name)] = struct{}{}
 		}
 	}
 
@@ -1039,7 +1224,7 @@ func (a Shards) FieldDimensions(measurements []string) (fields map[string]influx
 			return nil, nil, err
 		}
 		for k, typ := range f {
-			if _, ok := fields[k]; typ != influxql.Unknown && (!ok || typ < fields[k]) {
+			if fields[k].LessThan(typ) {
 				fields[k] = typ
 			}
 		}
@@ -1053,19 +1238,21 @@ func (a Shards) FieldDimensions(measurements []string) (fields map[string]influx
 func (a Shards) MapType(measurement, field string) influxql.DataType {
 	var typ influxql.DataType
 	for _, sh := range a {
-		if t := sh.MapType(measurement, field); typ.LessThan(t) {
+		sh.mu.RLock()
+		if t, err := sh.mapType(measurement, field); err == nil && typ.LessThan(t) {
 			typ = t
 		}
+		sh.mu.RUnlock()
 	}
 	return typ
 }
 
-func (a Shards) CreateIterator(measurement string, opt influxql.IteratorOptions) (influxql.Iterator, error) {
-	itrs := make([]influxql.Iterator, 0, len(a))
+func (a Shards) CreateIterator(ctx context.Context, measurement *influxql.Measurement, opt query.IteratorOptions) (query.Iterator, error) {
+	itrs := make([]query.Iterator, 0, len(a))
 	for _, sh := range a {
-		itr, err := sh.CreateIterator(measurement, opt)
+		itr, err := sh.CreateIterator(ctx, measurement, opt)
 		if err != nil {
-			influxql.Iterators(itrs).Close()
+			query.Iterators(itrs).Close()
 			return nil, err
 		} else if itr == nil {
 			continue
@@ -1074,8 +1261,8 @@ func (a Shards) CreateIterator(measurement string, opt influxql.IteratorOptions)
 
 		select {
 		case <-opt.InterruptCh:
-			influxql.Iterators(itrs).Close()
-			return nil, err
+			query.Iterators(itrs).Close()
+			return nil, query.ErrQueryInterrupted
 		default:
 		}
 
@@ -1083,12 +1270,63 @@ func (a Shards) CreateIterator(measurement string, opt influxql.IteratorOptions)
 		if opt.MaxSeriesN > 0 {
 			stats := itr.Stats()
 			if stats.SeriesN > opt.MaxSeriesN {
-				influxql.Iterators(itrs).Close()
+				query.Iterators(itrs).Close()
 				return nil, fmt.Errorf("max-select-series limit exceeded: (%d/%d)", stats.SeriesN, opt.MaxSeriesN)
 			}
 		}
 	}
-	return influxql.Iterators(itrs).Merge(opt)
+	return query.Iterators(itrs).Merge(opt)
+}
+
+func (a Shards) IteratorCost(measurement string, opt query.IteratorOptions) (query.IteratorCost, error) {
+	var costs query.IteratorCost
+	var costerr error
+	var mu sync.RWMutex
+
+	setErr := func(err error) {
+		mu.Lock()
+		defer mu.Unlock()
+		if costerr == nil {
+			costerr = err
+		}
+	}
+
+	limit := limiter.NewFixed(runtime.GOMAXPROCS(0))
+	var wg sync.WaitGroup
+	for _, sh := range a {
+		limit.Take()
+		wg.Add(1)
+
+		mu.RLock()
+		if costerr != nil {
+			mu.RUnlock()
+			break
+		}
+		mu.RUnlock()
+
+		go func(sh *Shard) {
+			defer limit.Release()
+			defer wg.Done()
+
+			engine, err := sh.engine()
+			if err != nil {
+				setErr(err)
+				return
+			}
+
+			cost, err := engine.IteratorCost(measurement, opt)
+			if err != nil {
+				setErr(err)
+				return
+			}
+
+			mu.Lock()
+			costs = costs.Combine(cost)
+			mu.Unlock()
+		}(sh)
+	}
+	wg.Wait()
+	return costs, costerr
 }
 
 func (a Shards) ExpandSources(sources influxql.Sources) (influxql.Sources, error) {
@@ -1097,7 +1335,9 @@ func (a Shards) ExpandSources(sources influxql.Sources) (influxql.Sources, error
 
 	// Iterate through every shard and expand the sources.
 	for _, sh := range a {
-		expanded, err := sh.ExpandSources(sources)
+		sh.mu.RLock()
+		expanded, err := sh.expandSources(sources)
+		sh.mu.RUnlock()
 		if err != nil {
 			return nil, err
 		}
@@ -1139,6 +1379,18 @@ func NewMeasurementFields() *MeasurementFields {
 	return &MeasurementFields{fields: make(map[string]*Field)}
 }
 
+func (m *MeasurementFields) FieldKeys() []string {
+	m.mu.RLock()
+	defer m.mu.RUnlock()
+
+	a := make([]string, 0, len(m.fields))
+	for key := range m.fields {
+		a = append(a, key)
+	}
+	sort.Strings(a)
+	return a
+}
+
 // MarshalBinary encodes the object to a binary format.
 func (m *MeasurementFields) MarshalBinary() ([]byte, error) {
 	m.mu.RLock()
@@ -1338,11 +1590,13 @@ type Field struct {
 
 // NewFieldKeysIterator returns an iterator that can be iterated over to
 // retrieve field keys.
-func NewFieldKeysIterator(sh *Shard, opt influxql.IteratorOptions) (influxql.Iterator, error) {
-	itr := &fieldKeysIterator{sh: sh}
+func NewFieldKeysIterator(engine Engine, opt query.IteratorOptions) (query.Iterator, error) {
+	itr := &fieldKeysIterator{engine: engine}
 
 	// Retrieve measurements from shard. Filter if condition specified.
-	names, err := sh.engine.MeasurementNamesByExpr(opt.Condition)
+	//
+	// FGA is currently not supported when retrieving field keys.
+	names, err := engine.MeasurementNamesByExpr(query.OpenAuthorizer, opt.Condition)
 	if err != nil {
 		return nil, err
 	}
@@ -1353,22 +1607,22 @@ func NewFieldKeysIterator(sh *Shard, opt influxql.IteratorOptions) (influxql.Ite
 
 // fieldKeysIterator iterates over measurements and gets field keys from each measurement.
 type fieldKeysIterator struct {
-	sh    *Shard
-	names [][]byte // remaining measurement names
-	buf   struct {
+	engine Engine
+	names  [][]byte // remaining measurement names
+	buf    struct {
 		name   []byte  // current measurement name
 		fields []Field // current measurement's fields
 	}
 }
 
 // Stats returns stats about the points processed.
-func (itr *fieldKeysIterator) Stats() influxql.IteratorStats { return influxql.IteratorStats{} }
+func (itr *fieldKeysIterator) Stats() query.IteratorStats { return query.IteratorStats{} }
 
 // Close closes the iterator.
 func (itr *fieldKeysIterator) Close() error { return nil }
 
 // Next emits the next tag key name.
-func (itr *fieldKeysIterator) Next() (*influxql.FloatPoint, error) {
+func (itr *fieldKeysIterator) Next() (*query.FloatPoint, error) {
 	for {
 		// If there are no more keys then move to the next measurements.
 		if len(itr.buf.fields) == 0 {
@@ -1377,7 +1631,7 @@ func (itr *fieldKeysIterator) Next() (*influxql.FloatPoint, error) {
 			}
 
 			itr.buf.name = itr.names[0]
-			mf := itr.sh.engine.MeasurementFields(itr.buf.name)
+			mf := itr.engine.MeasurementFields(itr.buf.name)
 			if mf != nil {
 				fset := mf.FieldSet()
 				if len(fset) == 0 {
@@ -1402,7 +1656,7 @@ func (itr *fieldKeysIterator) Next() (*influxql.FloatPoint, error) {
 
 		// Return next key.
 		field := itr.buf.fields[0]
-		p := &influxql.FloatPoint{
+		p := &query.FloatPoint{
 			Name: string(itr.buf.name),
 			Aux:  []interface{}{field.Name, field.Type.String()},
 		}
@@ -1413,10 +1667,10 @@ func (itr *fieldKeysIterator) Next() (*influxql.FloatPoint, error) {
 }
 
 // NewTagKeysIterator returns a new instance of TagKeysIterator.
-func NewTagKeysIterator(sh *Shard, opt influxql.IteratorOptions) (influxql.Iterator, error) {
+func NewTagKeysIterator(engine Engine, opt query.IteratorOptions) (query.Iterator, error) {
 	fn := func(name []byte) ([][]byte, error) {
 		var keys [][]byte
-		if err := sh.engine.ForEachMeasurementTagKey(name, func(key []byte) error {
+		if err := engine.ForEachMeasurementTagKey(name, func(key []byte) error {
 			keys = append(keys, key)
 			return nil
 		}); err != nil {
@@ -1424,16 +1678,15 @@ func NewTagKeysIterator(sh *Shard, opt influxql.IteratorOptions) (influxql.Itera
 		}
 		return keys, nil
 	}
-	return newMeasurementKeysIterator(sh, fn, opt)
+	return newMeasurementKeysIterator(engine, fn, opt)
 }
 
 // measurementKeyFunc is the function called by measurementKeysIterator.
 type measurementKeyFunc func(name []byte) ([][]byte, error)
 
-func newMeasurementKeysIterator(sh *Shard, fn measurementKeyFunc, opt influxql.IteratorOptions) (*measurementKeysIterator, error) {
+func newMeasurementKeysIterator(engine Engine, fn measurementKeyFunc, opt query.IteratorOptions) (*measurementKeysIterator, error) {
 	itr := &measurementKeysIterator{fn: fn}
-
-	names, err := sh.engine.MeasurementNamesByExpr(opt.Condition)
+	names, err := engine.MeasurementNamesByExpr(opt.Authorizer, opt.Condition)
 	if err != nil {
 		return nil, err
 	}
@@ -1453,13 +1706,13 @@ type measurementKeysIterator struct {
 }
 
 // Stats returns stats about the points processed.
-func (itr *measurementKeysIterator) Stats() influxql.IteratorStats { return influxql.IteratorStats{} }
+func (itr *measurementKeysIterator) Stats() query.IteratorStats { return query.IteratorStats{} }
 
 // Close closes the iterator.
 func (itr *measurementKeysIterator) Close() error { return nil }
 
 // Next emits the next tag key name.
-func (itr *measurementKeysIterator) Next() (*influxql.FloatPoint, error) {
+func (itr *measurementKeysIterator) Next() (*query.FloatPoint, error) {
 	for {
 		// If there are no more keys then move to the next measurements.
 		if len(itr.buf.keys) == 0 {
@@ -1478,7 +1731,7 @@ func (itr *measurementKeysIterator) Next() (*influxql.FloatPoint, error) {
 		}
 
 		// Return next key.
-		p := &influxql.FloatPoint{
+		p := &query.FloatPoint{
 			Name: string(itr.buf.name),
 			Aux:  []interface{}{string(itr.buf.keys[0])},
 		}
